
# THIS FILE IS GENERATED FROM SETUP.PY
short_version = '0.1.4'
version = '0.1.4'
full_version = '0.1.4'
git_revision = '1787df7398bf2b384288f2a7d3d9a7da87044dea'
release = True

if not release:
    version = full_version
